/*
 * Copyright (C) 2000-2006 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

/**
 * \file bt_psm.h
 * 
 * \author Kevin Martin <kevmarti@tik.ee.ethz.ch> 
 * 
 * \date 2006/03/23
 * 
 * \brief Bluetooth protocol/service multiplexor
 * 
 * The protocol/service multiplexor is the entity that delivers incoming
 * packets to different application services.
 * 
 * \par Service Registration
 * 
 * Services on the receiving side are identified by a "PSM", a unique
 * number which is included in every data packet. Thus, to receive data,
 * applications have to register a service routine that processes incoming
 * data, together with a "PSM" identifier at the protocol/service multiplexor.
 * See #bt_psm_service_register for further details.
 * 
 * \par Packet Processing
 * 
 * Incoming packets are encapsulated in ACL packet buffers #bt_acl_pkt_buf.
 * For every received buffer, a free buffer has to be passed back to
 * the protocol/service multiplexor for that it does not run out of
 * free buffers. This has to be done whenever a service rotine exits,
 * see #BT_PSM_DATA_CB.
 * 
 * \see
 * bt_acl_pkt_buf.h \n
 * l2cap_cl.h \n
 * mhop_cl.h
 */
#ifndef _BT_PSM_H_
#define _BT_PSM_H_

#include <bt/bt_acl_pkt_buf.h>

/**
 * \name Errors
 */
//@{
/** The psm specified is not valid */
#define BT_PSM_ERR_INVALID_PSM				-1

/** The psm specified is used by another registered service */
#define BT_PSM_ERR_PSM_IN_USE				-2

/** No more services can be registered to the connection-less data channel */
#define BT_PSM_ERR_SERVICE_PORTS_EXHAUSTED	-3

/** The service number passed does not identify a registered service */
#define BT_PSM_ERR_INVALID_SERVICE_NR		-5

/** The psm field in the data packet received is valid but there is no
 * service registered on that psm */
#define BT_PSM_ERR_NO_SERVICE_REGISTERED	-6

/** The service to deliver does not provide enough free buffers */
#define BT_PSM_ERR_BUF_OVERFLOW				-7
//@}

/**
 * \brief Service Routine Definition.
 *
 * A service routine handles incoming data and has to be specified
 * by the application during service registration
 * (see #bt_psm_service_register). This defines how a service
 * routine has to look like.
 * 
 * \param msg Pointer to the received ACL packet buffer (incoming
 * packets are encapsulated in ACL packet buffers)
 * \param data Pointer to the payload of the packet
 * \param len Size of the packet's payload
 * \param service_nr Number of the service this callback corresponds to
 * \param arg Pointer to the (optional) callback argument.
 * 
 * \returns Pointer to a free ACL packet buffer. Usually, just pass the
 * message obtained.
 * 
 * Note: Keep execution time of your service routines as short as
 * possible for that other services are not blocked too long.
 * I.e. try to avoid #NutSleep- and #NutEventWait statements,
 * as well as unnecessary thread-switches.
 * 
 * \see
 * #bt_psm_service_register \n
 * #bt_acl_pkt_buf_t \n
 * #bt_acl_pkt_buf
 */
#define BT_PSM_DATA_CB \
	bt_acl_pkt_buf* (*data_cb) (bt_acl_pkt_buf* pkt_buf, \
								u_char* data, \
								u_short len, \
								u_short service_nr, \
								void* cb_arg)

/**
 * \brief Protocol / service multiplexor typedef.
 */
typedef struct bt_psm_s bt_psm_t;


/**
 * \brief Returns the number of the free ACL packet buffers available
 *
 * \see #bt_psm_service_set_buffers
 */
inline u_short bt_psm_get_nr_default_bufs(bt_psm_t* psmux);


/**
 * \name Initialization
 */
//@{
/**
 * Initializes a protocol / service multiplexor. 
 * 
 * Note: The bt-stack, and the ACL layer, have to be initialized before
 * calling this function. Thus, make shure that:
 * - #bt_hci_init
 * - #bt_acl_init
 * have been called.
 * 
 * \param bt_stack Pointer to the bt-stack to use
 * \param nr_services Defines the maximal number of services the
 * multiplexor can handle.
 * \param nr_bufs Number of packet buffers to use
 * 
 * \return Pointer to the created multiplexor
 */
bt_psm_t* bt_psm_init(struct btstack* bt_stack,
						u_short nr_services,
						u_short nr_bufs);
//@}


/**
 * \name Application Interface
 */
//@{

/**
 * \brief Registers a higher layer service at the
 * protocol / service multiplexor.
 * 
 * After calling this function, packets received on PSM \p psm
 * are passed to the registering service. That is, each time a packet
 * is received, the registered service is signalled by calling the
 * service routine \p #BT_PSM_DATA_CB.
 * 
 * \param psmux Pointer to the multiplexor to use
 * \param psm The PSM to register on
 * \param #BT_PSM_DATA_CB Data callback function to register.
 * \param cb_arg Callback argument
 * 
 * Example:
\verbatim
// PSM your service uses
#define SERVICE_PSM 0x1009

// data reception funtion of your service
_service_data_rcv_cb(...) {
    ...
}

int main(void) {
	bt_psm_t* psmux;
	long service_nr;
	
	...
	
	// initialize protocol/service multiplexor
	psmux = bt_psm_init(bt_stack, 6, 6);

	// register your service
	servic_nr =
		bt_psm_service_register(psmux, SERVICE_PSM, _service_data_rcv_cb, NULL);
	
	...
}
\endverbatim
 * Data can then be sent to a service by sending data to the corresponding
 * PSM using one of the connection-less transport protocols, e.g.
\verbatim
u_char data[12];
strcpy(data, "hello world");
l2cap_cl_send(data, <con_handle>, SERVICE_PSM, 12);
\endverbatim
 *
 * \see
 * #BT_PSM_DATA_CB \n
 * l2cap_cl.h \n
 * mhop_cl.h
 */
long bt_psm_service_register(bt_psm_t* psmux, u_short psm, BT_PSM_DATA_CB, void* cb_arg);

/**
 * \brief Sets the packet buffers a registered service shall use.
 * 
 * Call this function if your service needs extra buffers or if
 * you'd like to turn off packet buffering completely. If extra
 * buffers are defined, incoming messages are buffered as long as
 * free buffers can be found in \p free_bufs.
 * 
 * \param mux Pointer to the protocol/service multiplexor
 * \param servic_nr Service number of the registered service
 * \param free_bufs Pointer to an initialized packet buffer queue
 * (see #bt_acl_pkt_queue_create). Pass NULL if buffering shall
 * be disabled.
 * 
 * Example:
\verbatim
// PSM your service uses
#define SERVICE_PSM 0x1009

// pointer to extra buffers
bt_acl_pkt_buf* bufs;

// data reception funtion of your service
bt_msg_t* _service_data_rcv_cb(...) {
    ...
}

int main(void) {
	bt_psm_t* psmux;
	long service_nr;
	
	...
	
	// initialize protocol/service multiplexor
	psmux = bt_psm_init(bt_stack, 6, 6);

	// register your service
	servic_nr =
		bt_psm_service_register(psmux, SERVICE_PSM, _service_data_rcv_cb, NULL);
	// create extra buffers
	bufs = bt_acl_pkt_que_create(6);
	bt_psm_service_set_buffers(psmux, service_nr, bufs);
	
	...
}
\endverbatim
 */
short bt_psm_service_set_buffers(bt_psm_t* mux,
								u_short service_nr,
								bt_acl_pkt_buf** free_bufs);

/** \brief Clears a previousely registered service
 * 
 * Clears the registered service with service nr \p service_nr.
 * 
 * Note: dedicated buffers beloning to the service to be cleared are
 * not destroyed.
 * 
 * \param stack Pointer to the connection-less l2cap stack
 * \param service_nr Service number of the service to clear
 */
short bt_psm_service_clear(bt_psm_t* mux, u_short service_nr);
//@}


/** \brief Delivers a received message to a registered service
 * 
 * Delivers the received message \p msg to the service registered at
 * PSM \p psm. The parameters \p data and \p len are passed
 * to the higher layer service via the registered data callback
 * (see #BT_PSM_DATA_CB). Thus, \p data should be a pointer to the payload
 * of the packet, while \p data_len should be set to the lenght of the
 * payload.
 * 
 * Note: applications typically do not call this function! This function
 * is used by lower layer protocols (e.g. l2cap connection-less)
 * to deliver received messages to higher layer protocols / services
 * (e.g. multi-hop layer, application service).
 * 
 * \param mux Pointer to the protocol / service multiplexor
 * \param msg Pointer to the message to be delivered
 * \param data Pointer to the payload of the message
 * \param len Size of the payload
 * \param psm The PSM the message \p msg will be delivered to
 * \param free_msg Reference to a free packet buffer pointer:
 * the protocol / service multiplexor will store a pointer
 * to a free message buffer in the reference passed. If there is no
 * free packet buffer available, the buffer \p msg passed will be
 * returned.
 * 
 * \returns 0 or possible errors:
 * - BT_PSM_ERR_INVALID_PSM
 * - BT_PSM_ERR_NO_SERVICE_REGISTERED
 * - BT_PSM_ERR_BUF_OVERFLOW
 */
short bt_psm_deliver_msg(bt_psm_t* mux,
						bt_acl_pkt_buf* pkt_buf,
						u_char* data,
						u_short len,
						u_short psm,
						bt_acl_pkt_buf** free_buf_p);

#endif //_BT_PSM_H_
