/*
 * Copyright (C) 2000-2004 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

 /*
 * btn-eepromdb.h
 *
 * 02.07.2004 Martin Hinz <btnode@hinz.ch>
 */

/**
 *
 * \file btn-eepromdb.h
 * \author Martin Hinz <btnode@hinz.ch>
 * 
 * \brief Provides support for storing and retrieving 'features' of
 * the particular BTnode the software is running. This can be used to
 * store, how much memory is availabe, or what the user readable name
 * of this device is. Or what sensors are connected to it.
 * 
 * Features are addressed using a four byte identifier 'xxxx'. For
 * string features, the last byte has to be '=', so all string feature
 * id look like 'xxx=' followed by the string in the eeprom.
 *
 * The feature data is stored in the EEPROM of the device. The first
 * implementation supports Int32 and strings (len <= 16) feature data.
 *
 * To create such a feature ID, you can use the macro FOURCC(a,b,c,d)
 * that takes four characters and creates a 4 char constant. The simple
 * use of 'abcd' is deprecated.
 *
 */
#ifndef EEPROMDB_H_INCLUDED
#define EEPROMDB_H_INCLUDED

#include <io.h>

#define BTN_EEPROMDB_ID_SIZE 4
#define BTN_EEPROMDB_INT_SIZE  4
#define BTN_EEPROMDB_STRING_SIZE  16

/*!
  
\brief Create a 32 constant from 4 single characters

\param a,b,c,d id characters
\return 32 bit integer constant
*/

#define FOURCC(a,b,c,d) ( (u_long) a << 24 | (u_long) b << 16 | (u_long) c << 8 | (u_long) d )

/*!
  
\brief Stores into the eepromdb on the EEPROM of the BTnode

\param id 32 bit Feature id. String id have to be 'xxx='"
\param size Size of the data to store. (Only relevant with string values, 
            maximal #BTN_EEPROMDB_STRING_SIZE bytes
\param data Pointer the data to be stored (string or pointer of long)
\return Size of the data successfully stored or 0 to indicate an error.
*/

u_short btn_eepromdb_store(u_long id, u_short size, void *data);

/*!
  
\brief Get a value in the eeprom database in the EEPROM of the BTnode

\param id 32 bit value id. String value have to be 'xxx='"
\param size Max size of the data buffer. Only used for strings.
\param data Pointer the data buffer
\return Size of the data retrieved (strlen in case of string) 
        or 0 to indicate an error.
\note: To retrieve a string with len=16, a buffer of size=17 is needed
       as the terminating '\0' is always added.
*/

u_short btn_eepromdb_get(u_long id, u_short size, void *data);

/*!
  
\brief Delete a value in the eeprom database in the EEPROM of the BTnode

\param id 32 bit value id. String features have to be 'xxx='"
\return 1 ok or 0 to indicate an error.
*/

u_char btn_eepromdb_delete(u_long id);

/*!
  
\brief Get the value id of value nr index. This is used to enumerate the eeprom database

\param index Index of value id to get
\return Value ID or 0 for last value
*/

u_long btn_eepromdb_get_by_index(u_short index);

#endif

